function [Q, S, N, s_maps, s_local] = TMQI_NSS_Ent(hdrImage, ldrImage, window)
% ========================================================================
% Tone Mapped image Quality Index (TMQI), Version 1.0
% Copyright(c) 2012 Hojatollah Yeganeh and Zhou Wang
%
% Tone Mapped image Quality Index using MSCN based natural scene statistics
% and entropy based pooling (TMQI-NSS-Ent)
% Copyright(c) 2016 Debarati Kundu and Brian Evans
% All Rights Reserved.
%
% ----------------------------------------------------------------------
% Permission to use, copy, or modify this software and its documentation
% for educational and research purposes only and without fee is hereby
% granted, provided that this copyright notice and the original authors'
% names appear on all copies and supporting documentation. This program
% shall not be used, rewritten, or adapted as the basis of a commercial
% software or hardware product without first obtaining permission of the
% authors. The authors make no representations about the suitability of
% this software for any purpose. It is provided "as is" without express
% or implied warranty.
%----------------------------------------------------------------------
% This code builds on top of the original TMQI metric and uses a pooling
% measure based on the local entropy of the image.
%
% Please refer to the following paper and the website with suggested usage
%
% H. Yeganeh and Z. Wang, "Objective Quality Assessment of Tone Mapped
% Images," IEEE Transactios on Image Processing, vol. 22, no. 2, pp. 657- 
% 667, Feb. 2013.
%
% Kundu, D.; Evans, B.L.;, “Visual Attention guided Quality Assessment of 
% Tone-Mapped images using Scene Statistics,” submitted to IEEE International 
% Conference on Image Processing, Sept. 2016
%
%
% Kindly report any suggestions or corrections to debarati@utexas.edu
%
%----------------------------------------------------------------------
%
%Input : (1) hdrImage: the HDR image being used as reference. The HDR 
%            image must be an m-by-n-by-3 double array
%        (2) queryImage: the query image being compared (may be LDR image in
%                      case of tone-mapped image or another HDR image in case of 
%                      joint tone-mapped and JPEG compressed images)
%        (3) window: local window for statistics (see the above
%            reference). default widnow is Gaussian given by
%            window = fspecial('gaussian', 11, 1.5);
%
%Output: (1) s_maps: The structural fidelity maps of the query image.     
%        (2) s_local: the mean of s_map in each scale (see above refernce).
%        (3) S: The structural fidelity score of the query test image. 
%        (4) N: The statistical naturalness score of the query image.
%        (5) Q: The TMQI score of the query image. 
%
%Basic Usage:
%   Given query image and its corresponding HDR images, 
%
%   [Q, S, N, s_maps, s_local] = TMQI_NSS_Ent(hdrImage, queryImage);
%
%Advanced Usage:
%   User defined parameters. For example
%   window = ones(8);
%   [Q, S, N, s_maps, s_local] = TMQI_NSS_Ent(hdrImage, queryImage, window);
%
%========================================================================

if (nargin < 2 || nargin > 3)
   s_maps = -Inf;
   s_local = -Inf;
   S = -Inf;
   N = -Inf;
   Q = -Inf;
   return;
end

if (size(hdrImage) ~= size(ldrImage))
   s_maps = -Inf;
   s_local = -Inf;
   S = -Inf;
   N = -Inf;
   Q = -Inf;
   return;
end

[M,N,D] = size(hdrImage);

if (nargin == 2)
   if ((M < 11) || (N < 11))
	   s_maps = -Inf;
       s_local = -Inf;
       S = -Inf;
       N = -Inf;
       Q = -Inf;
       disp('the image size is less than the window size'); 
     return
   end
   window = fspecial('gaussian', 11, 1.5);	%
end

if (nargin == 3)
   [H, W] = size(window);
   if ((H*W) < 4 || (H > M) || (W > N))
	   s_maps = -Inf;
       s_local = -Inf;
       S = -Inf;
       N = -Inf;
       Q = -Inf;
      return
   end
end


%---------- default parameters -----%
a = 0.8012;
Alpha = 0.3046;
Beta = 0.7088;

%---------- default parameters -----
level = 5;
weight = [0.0448 0.2856 0.3001 0.2363 0.1333];
%--------------------
HDR = RGBtoYxy(hdrImage);
L_hdr = HDR(:,:,1);
if (isa(hdrImage,'double'))
	lmin = min(min(L_hdr));
	lmax = max(max(L_hdr));
	L_hdr = double(round((2^32 - 1)/(lmax - lmin)).*(L_hdr - lmin));
end
%-------------------------------------------
LDR = RGBtoYxy(ldrImage);
L_ldr = LDR(:,:,1);
if (isa(ldrImage,'double'))
	lmin = min(min(L_ldr));
	lmax = max(max(L_ldr));
	L_ldr = double(round((2^32 - 1)/(lmax - lmin)).*(L_ldr - lmin));
end
%----------- structural fidelity -----------------
[S, s_local, s_maps] = StructuralFidelity(L_hdr, L_ldr,level,weight, window);
%--------- statistical naturalness ---------------
N = StatisticalNaturalnessMSCN(L_ldr);
%------------- overall quality -----------------
Q = a*(S^Alpha) + 0.5*(1-a)*(N(1)^Beta) + 0.5*(1-a)*(N(2)^Beta); 
end

%================ Color Space Conversion ==============================

function [Yxy] = RGBtoYxy(RGB)
RGB = double(RGB);
M = [ 0.4124 0.3576 0.1805
      0.2126 0.7152 0.0722
      0.0193  0.1192 0.9505];

 RGB2 = reshape(RGB,size(RGB,1)*size(RGB,2),3);
 XYZ2 = M *  RGB2';
 S = sum(XYZ2);
 Yxy2 = zeros(size(XYZ2));
 Yxy2(1,:)=XYZ2(2,:);
 Yxy2(2,:)=XYZ2(1,:)./(S+eps);
 Yxy2(3,:)=XYZ2(2,:)./(S+eps);
 Yxy = reshape(Yxy2',size(RGB,1),size(RGB,2),3);
end

%================= Structural Fidelity Measure =========================

function [S, s_local, s_maps] = StructuralFidelity(L_hdr, L_ldr,level,weight, window)
downsample_filter = ones(2)./4;
L_hdr = double(L_hdr);
L_ldr = double(L_ldr);
f = 32;
for l = 1:level
    f = f/2;
    [s_local(l), smap] = Slocal(L_hdr, L_ldr, window , f);
    s_maps{l} = smap;
    filtered_im1 = imfilter(L_hdr, downsample_filter, 'symmetric', 'same');
    filtered_im2 = imfilter(L_ldr, downsample_filter, 'symmetric', 'same');
    clear L_hdr;
    clear L_ldr;
    L_hdr = filtered_im1(1:2:end, 1:2:end);
    L_ldr = filtered_im2(1:2:end, 1:2:end);
end
S = prod(s_local.^weight(1:level));
end

%============ Local Structural Fidelity Measure =========================

function [s, s_map] = Slocal(L_hdr, L_ldr,window, sf)
C1 = 0.01;
C2 = 10;
window = window/sum(sum(window));
%----------------------------------------------------
minHeight = min(size(L_hdr,1),size(L_ldr,1));
minWidth = min(size(L_hdr,2),size(L_ldr,2));
L_hdr = L_hdr(1:minHeight, 1:minWidth);
L_ldr = L_ldr(1:minHeight, 1:minWidth);

img1 = double(L_hdr);
img2 = double(L_ldr);
mu1 = filter2(window, img1, 'valid');
mu2 = filter2(window, img2, 'valid');
mu1_sq = mu1.*mu1;
mu2_sq = mu2.*mu2;
mu1_mu2 = mu1.*mu2;
sigma1_sq = filter2(window, img1.*img1, 'valid') - mu1_sq;
sigma2_sq = filter2(window, img2.*img2, 'valid') - mu2_sq;
sigma1 = sqrt(max(0, sigma1_sq));
sigma2 = sqrt(max(0, sigma2_sq));
sigma12 = filter2(window, img1.*img2, 'valid') - mu1_mu2;
%-------------------------------------------------------
% Mannos CSF Function
CSF = 100.*2.6*(0.0192+0.114*sf)*exp(-(0.114*sf)^1.1);
%-----------------------
u_hdr= 128/(1.4*CSF); 
sig_hdr=u_hdr/3;
sigma1p = normcdf(sigma1,u_hdr,sig_hdr);
%---------------------------------------------------------
u_ldr = u_hdr;
sig_ldr = u_ldr/3;
sigma2p = normcdf(sigma2,u_ldr,sig_ldr);
%----------------------------------------------------
s_map = (((2*sigma1p.*sigma2p)+C1)./((sigma1p.*sigma1p)+(sigma2p.*sigma2p)+C1)).*((sigma12+C2)./(sigma1.*sigma2 + C2));
weight = entropyfilt(uint8(img2),true(size(window,1))); 
edgeSize = (size(window,1)-1)/2;
weight = weight(edgeSize + 1:size(weight,1)-edgeSize, edgeSize + 1:size(weight,2)-edgeSize);
weight = weight./(sum(sum(weight)));
s_map = s_map.*abs(weight);
s = mean2(s_map);
end

%============ Statistical Naturalness Measure ================

function [N] = StatisticalNaturalnessMSCN(L_ldr)
im = L_ldr;
window = fspecial('gaussian',7,7/6);
window = window/sum(sum(window));
mu = imfilter(L_ldr,window,'replicate');
mu_sq = mu.*mu;
sigma = sqrt(abs(imfilter(im.*im,window,'replicate') - mu_sq));
structdis = (im-mu)./(sigma+1);  
[alpha, overallstd] = estimateggdparam(structdis(:));
N = [alpha, overallstd, std(sigma(:))];
end
